import argparse
import logging
import time
from logging import error, info
from pathlib import Path
from typing import Optional

from watchdog.events import EVENT_TYPE_CLOSED, FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

try:
    from watchdog.events import EVENT_TYPE_OPENED  # pylint: disable=ungrouped-imports
except ImportError:
    EVENT_TYPE_OPENED = "opened"  # type: ignore

from tumfl import ParserError, format, minifier
from tumfl.AST import ASTNode
from tumfl.config import Config, parse_config
from tumfl.dependency_resolver import resolve_recursive
from tumfl.error import TumflError
from tumfl.formatter import MinifiedStyle


class RunConfig:
    # pylint: disable=too-few-public-methods
    def __init__(
        self, source: Path, destination: Optional[Path], config: Config, minify: bool
    ) -> None:
        self.source: Path = source
        self.destination: Optional[Path] = destination
        self.config: Config = config
        self.minify: bool = minify


def compile_file(filename: Path, config: Config, minify: bool) -> str:
    source_directory: Path = filename.parent
    result: ASTNode = resolve_recursive(
        filename, [source_directory], add_source_description=True, config=config
    )
    if minify:
        minifier.minify(result)
        return format(result, MinifiedStyle)
    return format(result)


def run(config: RunConfig) -> None:
    if not config.destination:
        config.destination = config.source.with_stem(config.source.stem + "_result")
    try:
        compiled = compile_file(config.source, config.config, config.minify)
    except RuntimeError:
        return
    except TumflError as e:
        if isinstance(e, ParserError):
            error(e.full_error)
        else:
            error(e)
        return
    with config.destination.open("w") as f:
        f.write(compiled)


class EventHandler(FileSystemEventHandler):
    def __init__(self, config: RunConfig):
        self.config: RunConfig = config

    def execute(self) -> None:
        info("Files changed, recompiling")
        run(self.config)

    def on_any_event(self, event: FileSystemEvent) -> None:
        if event.event_type not in (EVENT_TYPE_CLOSED, EVENT_TYPE_OPENED):
            self.execute()


def follow(config: RunConfig) -> None:
    run(config)
    observer = Observer()
    event_handler = EventHandler(config)
    observer.schedule(event_handler, str(config.source.parent), recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(100)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def main() -> None:
    parser = argparse.ArgumentParser(description="Compile lua files", prog="tumfl")
    parser.add_argument("source_file", type=Path, help="Source file to compile")
    parser.add_argument("-d", "--destination-file", type=Path, help="Destination file")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Be verbose",
        action="count",
        default=0,
    )
    parser.add_argument(
        "-f", "--follow", action="store_true", help="Follow file changes"
    )
    parser.add_argument(
        "-c",
        "--config-file",
        type=Path,
        help="Replace placeholders using config file."
        " Config file is a normal lua file, with top level assignments of `name = value`, not limited to strings."
        " To use an replacement, just use a string int the target file like"
        ' `"$$name"` (if `$$` is your prefix, and `name` the name to look up).',
    )
    parser.add_argument(
        "--config-prefix",
        default="$$",
        help="Prefix for names to be replaced by config values (default: $$)",
    )
    parser.add_argument(
        "-m",
        "--minify",
        action="store_true",
        help="Minify the output",
    )
    args = parser.parse_args()
    loglevel = logging.WARN
    if args.verbose == 1:
        loglevel = logging.INFO
    elif args.verbose > 1:
        loglevel = logging.DEBUG
    logging.basicConfig(
        format="%(asctime)s|%(levelname)s| %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        level=loglevel,
    )
    config = Config({})
    if args.config_file:
        config = parse_config(args.config_file)
    config.prefix = args.config_prefix
    run_config = RunConfig(args.source_file, args.destination_file, config, args.minify)
    if args.follow:
        follow(run_config)
    else:
        run(run_config)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence

from .block import Block

if TYPE_CHECKING:
    from tumfl.AST import Expression, Statement
    from tumfl.Token import Token


class Chunk(Block):
    """Same as block, but individually executable."""

    def __init__(
        self,
        token: Token,
        statements: Sequence[Statement],
        returns: Optional[Sequence[Expression]],
    ):
        super().__init__(token, statements, returns)
        self.name = "Chunk"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from .AST import (
    ASTNode,
    Block,
    Chunk,
    ExpFunctionCall,
    ExpFunctionDefinition,
    FunctionCall,
    Name,
    NumberedTableField,
    Semicolon,
    String,
    Table,
)
from .basic_walker import NoneWalker
from .error import InvalidDependencyError
from .parser import parse
from .Token import Token

if TYPE_CHECKING:
    from .config import Config


def _parse_file(path: Path, config: Optional[Config]) -> Chunk:
    with path.open() as f:
        chunk: str = f.read()
    ast: Chunk = parse(chunk)
    ast.parent(None, path)
    if config:
        config.visit(ast)
    return ast


class ResolveDependencies(NoneWalker):
    def __init__(
        self,
        search_path: list[Path],
        add_source_description: bool = False,
        config: Optional[Config] = None,
    ):
        self.search_path: list[Path] = search_path
        self.found: dict[Path, Optional[list[Name]]] = {}
        self.add_source_description: bool = add_source_description
        self.config: Optional[Config] = config

    def _find_file_in_path(self, name: str, start_path: Path) -> Optional[Path]:
        parts: list[str] = name.split(".")
        if not parts[0]:
            return None
        filename: Path = Path(parts[0])
        for file in parts[1:]:
            filename /= file
        for path in (start_path, *self.search_path):
            for suffix in ("", ".tl", ".lua"):
                current_path = path.joinpath(filename.with_suffix(suffix))
                if current_path.is_file():
                    return current_path
        return None

    def _get_block_dependency_path(
        self, name: str, start_path: Path, token: Token, deduplicate: bool = True
    ) -> Optional[Path]:
        path: Optional[Path] = self._find_file_in_path(name, start_path)
        if not path:
            raise InvalidDependencyError(
                f"Could not find dependency with name {name}", token
            )
        if deduplicate and path in self.found:
            return None
        self.found[path] = None
        return path

    def __find_and_parse(
        self, name: str, node: ASTNode, deduplicate: bool, results: list[Chunk]
    ) -> None:
        assert node.file_name
        name = name.replace("/", ".")
        dependency_path: Optional[Path] = self._get_block_dependency_path(
            name, node.file_name.parent, node.token, deduplicate
        )
        if dependency_path:
            ast: Chunk = _parse_file(dependency_path, self.config)
            if self.add_source_description:
                ast.comment.insert(0, f"Sourced from {dependency_path}")
            results.append(ast)

    def __get_table_dependencies(
        self, node: Table, results: list[Chunk], deduplicate: bool
    ) -> None:
        for item in node.fields:
            if isinstance(item, NumberedTableField):
                if isinstance(item.value, String):
                    self.__find_and_parse(item.value.value, node, deduplicate, results)
                else:
                    raise InvalidDependencyError(
                        f"Wrong require() arguments. Table can only contain strings, got {item}",
                        node.token,
                    )
            else:
                raise InvalidDependencyError(
                    f"Wrong require() arguments. Can only use strings with no explicit index, got {item}",
                    node.token,
                )

    def __get_ast(
        self, node: Union[FunctionCall, ExpFunctionCall], deduplicate: bool = True
    ) -> Optional[Chunk | Semicolon | Block]:
        if isinstance(node.function, Name) and node.function.variable_name == "require":
            results: list[Chunk] = []
            for arg in node.arguments:
                if isinstance(arg, String):
                    self.__find_and_parse(arg.value, node, deduplicate, results)
                elif isinstance(arg, Table):
                    self.__get_table_dependencies(arg, results, deduplicate)
                else:
                    raise InvalidDependencyError(
                        f"Wrong require() arguments. Expected string or table, got {arg}",
                        node.token,
                    )
            ast: Union[Chunk, Semicolon, Block]
            if not results:
                ast = Semicolon(node.token)
                ast.parent(node.parent_class, node.file_name)
            elif len(results) == 1:
                ast = results[0]
            else:
                ast = Block(node.token, results, None)
            return ast
        return None

    def visit_FunctionCall(self, node: FunctionCall) -> None:
        if ast := self.__get_ast(node):
            if isinstance(ast, Semicolon):
                assert node.parent_class
                node.remove()
            else:
                ast.parent_class = node.parent_class
                node.replace(ast)
                self.visit(ast)
        else:
            super().visit_FunctionCall(node)

    def visit_ExpFunctionCall(self, node: ExpFunctionCall) -> None:
        # if the body is expecting a return value, never deduplicate
        if ast := self.__get_ast(node, deduplicate=False):
            # since deduplication is off, ast will never be a Semicolon
            assert isinstance(ast, Block)
            function = ExpFunctionDefinition(node.token, [], ast)
            node.function = function
            ast.parent_class = function
            call = ExpFunctionCall(node.token, function, [])
            function.parent_class = call
            call.parent_class = node.parent_class
            ast.file_name = function.file_name = call.file_name = node.file_name
            node.replace(call)
            self.visit(call)
        else:
            super().visit_ExpFunctionCall(node)


def resolve_recursive(
    path: Path,
    search_path: list[Path],
    add_source_description: bool = False,
    config: Optional[Config] = None,
) -> ASTNode:
    ast = _parse_file(path, config)
    resolver = ResolveDependencies(search_path, add_source_description, config=config)
    resolver.visit(ast)
    return ast



from __future__ import annotations

from typing import TYPE_CHECKING

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import Block, Expression
    from tumfl.Token import Token


class While(Statement):
    """The while loop"""

    def __init__(self, token: Token, condition: Expression, body: Block):
        super().__init__(token, "While")
        self.condition: Expression = condition
        self.body: Block = body



from .ast_node import ASTNode
from .base_function_definition import BaseFunctionDefinition
from .expression import *
from .statement import *

# pylint: disable=duplicate-code
__all__ = [
    "ASTNode",
    "BaseFunctionDefinition",
    "BinaryOperand",
    "BinOp",
    "Boolean",
    "ExpFunctionCall",
    "ExpFunctionDefinition",
    "ExpMethodInvocation",
    "Expression",
    "Index",
    "Name",
    "NamedIndex",
    "Nil",
    "Number",
    "String",
    "Table",
    "ExplicitTableField",
    "NamedTableField",
    "NumberedTableField",
    "TableField",
    "UnaryOperand",
    "UnOp",
    "Vararg",
    "Variable",
    "Assign",
    "Block",
    "Break",
    "Chunk",
    "FunctionCall",
    "FunctionDefinition",
    "Goto",
    "If",
    "IterativeFor",
    "Label",
    "AttributedName",
    "LocalAssign",
    "LocalFunctionDefinition",
    "MethodInvocation",
    "NumericFor",
    "Repeat",
    "Semicolon",
    "Statement",
    "While",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from tumfl.Token import Token

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import Name


class Label(Statement):
    """A label to be used with goto"""

    def __init__(self, token: Token, label_name: Name):
        super().__init__(token, "Label")
        self.label_name: Name = label_name



from __future__ import annotations

from itertools import zip_longest
from typing import Optional

from tumfl.AST import Name
from tumfl.minifier.util.replacements import ReplacementCollection, key_function
from tumfl.minifier.util.variable import Variable


class Scope:
    def __init__(self, parent: Optional[Scope] = None) -> None:
        self.parent: Optional[Scope] = None
        self.root: Scope = self
        self.level: int = 0
        self.children: list[Scope] = []
        self.variables: dict[str, Variable] = {}
        if parent:
            self.add_parent(parent)

    def add_parent(self, parent: Scope) -> None:
        self.parent = parent
        parent.children.append(self)
        self.root = parent.root
        self.level = parent.level + 1

    def __register(self, node: Name) -> None:
        node.set_attribute(self)

    def add_variable(
        self, name: Name, preserve: bool, global_: bool = False, add_write: bool = True
    ) -> Variable:
        if global_:
            if name.variable_name in self.variables:
                var = self.variables[name.variable_name]
                var.add_write(name)
                return var
            if self.parent:
                return self.parent.add_variable(name, preserve, global_, add_write)
            return self.root.add_variable(name, preserve)
        self.variables[name.variable_name] = self.variables.get(
            name.variable_name, Variable(self.root is self, preserve)
        )
        self.variables[name.variable_name].add_scope(self, name)
        if add_write:
            self.variables[name.variable_name].add_write(name)
        self.__register(name)
        return self.variables[name.variable_name]

    def use_variable(self, name: Name, is_method: bool) -> Variable:
        if name.variable_name in self.variables:
            self.variables[name.variable_name].add_read(name, is_method)
            self.__register(name)
            return self.variables[name.variable_name]
        if self.parent:
            return self.parent.use_variable(name, is_method)
        # unknown global
        self.add_variable(name, False, False, False)
        return self.use_variable(name, is_method)

    def __repr__(self) -> str:
        return f"Scope({self.level=}, {self.variables=}, {self.children=})"

    def cleanup_scopes(self) -> None:
        for var in self.variables.values():
            var.cleanup()
        for child in range(len(self.children) - 1, -1, -1):
            self.children[child].cleanup_scopes()
        if len(self.variables) == 0 and len(self.children) == 0 and self.parent:
            self.parent.children.remove(self)

    def remove_var(self, name: str) -> None:
        del self.variables[name]

    def collect_replacements(self) -> ReplacementCollection:
        result: ReplacementCollection = []
        for var in self.variables.values():
            result.extend(var.collect_replacements())

        to_evaluate: list[ReplacementCollection] = []
        for child in self.children:
            to_evaluate.append(child.collect_replacements())
        result.extend(
            sum(group, []) for group in zip_longest(*to_evaluate, fillvalue=[])
        )

        result.sort(key=key_function, reverse=True)
        return result



from tumfl.AST import Assign, Block, LocalAssign, Name
from tumfl.basic_walker import NoneWalker


class Optimize(NoneWalker):
    def visit_Block(self, node: Block) -> None:
        super().visit_Block(node)
        local_assigns: list[LocalAssign] = [
            stmt for stmt in node.statements if isinstance(stmt, LocalAssign)
        ]
        if len(local_assigns) > 1:
            new_assign = LocalAssign(local_assigns[0].token, [], [])
            for assign in local_assigns:
                new_assign.variable_names.extend(assign.variable_names)
            node.statements.insert(0, new_assign)
            for assign in local_assigns:
                if assign.expressions:
                    targets: list[Name] = [name.name for name in assign.variable_names]
                    normal_assign = Assign(assign.token, targets, assign.expressions)
                    assign.replace(normal_assign)
                else:
                    assign.remove()



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from tumfl.Token import Token

from .expression import Expression

if TYPE_CHECKING:
    from tumfl.AST import TableField


class Table(Expression):
    """A lua table constructor, like {[1]=2, 3; b=4}"""

    def __init__(self, token: Token, fields: Sequence[TableField]):
        super().__init__(token, "Table")
        self.fields: list[TableField] = list(fields)



from tumfl.AST import Block, LocalAssign
from tumfl.basic_walker import NoneWalker


class CombineLocal(NoneWalker):
    def visit_Block(self, node: Block) -> None:
        for i in range(len(node.statements) - 1, 0, -1):
            if isinstance(node.statements[i], LocalAssign) and isinstance(
                node.statements[i - 1], LocalAssign
            ):
                to_remove = node.statements[i]
                assert isinstance(to_remove, LocalAssign)
                move_into = node.statements[i - 1]
                assert isinstance(move_into, LocalAssign)
                if (
                    to_remove.expressions is not None
                    and move_into.expressions is not None
                ):
                    move_into.variable_names.extend(to_remove.variable_names)
                    move_into.expressions.extend(to_remove.expressions)
                    node.statements.pop(i)
        super().visit_Block(node)



from __future__ import annotations

from logging import warning
from typing import Optional

from tumfl.AST import (
    ASTNode,
    BinaryOperand,
    BinOp,
    Block,
    Boolean,
    ExpFunctionCall,
    ExpFunctionDefinition,
    ExpMethodInvocation,
    FunctionCall,
    FunctionDefinition,
    If,
    LocalFunctionDefinition,
    MethodInvocation,
    Name,
    NamedIndex,
    NamedTableField,
    Semicolon,
    UnaryOperand,
    UnOp,
)
from tumfl.basic_walker import AggregatingWalker, NoneWalker

from ..to_lua_type import to_lua_type
from ..to_python_type import ToPythonType
from .util.remove_name import RemoveName
from .util.variable import Variable


class ReplaceName(NoneWalker):
    def __init__(self, replacements: dict[Name, Name]) -> None:
        super().__init__()
        self.replacements: dict[Name, Name] = replacements

    def visit_Name(self, node: Name) -> None:
        if node in self.replacements:
            parent: Optional[ASTNode] = node.parent_class
            if isinstance(parent, NamedIndex) and node is not parent.lhs:
                return
            if (
                isinstance(parent, (MethodInvocation, ExpMethodInvocation))
                and node is not parent.function
            ):
                return
            if isinstance(parent, NamedTableField) and node is not parent.value:
                return
            node.replace(self.replacements[node])


class HasReturn(AggregatingWalker[bool]):
    def aggregation_function(self, a: bool, b: bool) -> bool:
        return a or b

    def default_value(self) -> bool:
        return False

    def visit_Block(self, node: Block) -> bool:
        return bool(node.returns) or super().visit_Block(node)


class Simplify(NoneWalker):
    def __init__(self) -> None:
        super().__init__()
        self.has_return = HasReturn()
        self.remove_name = RemoveName()
        self.to_python = ToPythonType()

    def visit_BinOp(self, node: BinOp) -> None:
        super().visit_BinOp(node)
        if (result := self.to_python.visit(node)) is not None:
            as_lua = to_lua_type(result)
            node.replace(as_lua)

    def visit_UnOp(self, node: UnOp) -> None:
        super().visit_UnOp(node)
        if (result := self.to_python.visit(node)) is not None:
            as_lua = to_lua_type(result)
            node.replace(as_lua)
        elif node.op == UnaryOperand.NOT:
            if isinstance(node.right, BinOp):
                new_op: Optional[BinaryOperand] = None
                if node.right.op == BinaryOperand.EQUALS:
                    new_op = BinaryOperand.NOT_EQUALS
                elif node.right.op == BinaryOperand.NOT_EQUALS:
                    new_op = BinaryOperand.EQUALS
                elif node.right.op == BinaryOperand.GREATER_THAN:
                    new_op = BinaryOperand.LESS_EQUALS
                elif node.right.op == BinaryOperand.GREATER_EQUALS:
                    new_op = BinaryOperand.LESS_THAN
                elif node.right.op == BinaryOperand.LESS_THAN:
                    new_op = BinaryOperand.GREATER_EQUALS
                elif node.right.op == BinaryOperand.LESS_EQUALS:
                    new_op = BinaryOperand.GREATER_THAN
                if new_op is not None:
                    node.right.op = new_op
                    node.replace(node.right)
            elif isinstance(node.right, UnOp) and node.right.op == UnaryOperand.NOT:
                node.replace(node.right.right)
        elif node.op == UnaryOperand.MINUS:
            if isinstance(node.right, BinOp):
                if node.right.op in (BinaryOperand.PLUS, BinaryOperand.MINUS):
                    node.right.op = (
                        BinaryOperand.MINUS
                        if node.right.op == BinaryOperand.PLUS
                        else BinaryOperand.PLUS
                    )
                    node.replace(node.right)
            elif isinstance(node.right, UnOp):
                if node.right.op == UnaryOperand.MINUS:
                    node.replace(node.right.right)

    def visit_If(self, node: If) -> None:
        super().visit_If(node)
        if isinstance(node.test, Boolean):
            if node.test.value:
                node.replace(node.true)
            elif node.false:
                node.replace(node.false)
            else:
                node.remove()

    def visit_FunctionCall(self, node: FunctionCall) -> None:
        super().visit_FunctionCall(node)
        var = node.get_attribute(Variable)
        if var and len(var.writes) == 1:
            definition: Optional[ASTNode] = var.writes[0].parent_class
            if not isinstance(
                definition,
                (FunctionDefinition, LocalFunctionDefinition, ExpFunctionDefinition),
            ):
                warning(
                    f'Function call to "{node.function!s}" is not a function definition'
                )
                return
            if (
                all(
                    isinstance(child, Semicolon) for child in definition.body.statements
                )
                and not definition.body.returns
            ):
                self.remove_name(node)
                node.remove()
                if len(var.reads) == 0:
                    definition.remove()
            elif (
                len(var.reads) == 1
                and all(isinstance(arg, Name) for arg in node.arguments)
                and all(isinstance(arg, Name) for arg in definition.parameters)
                and not self.has_return(definition.body)
            ):
                # Inlining is quite conservative, inlining only functions that have only name arguments,
                # and no return statements (and for obvious reasons, only if there is only 1 callee)
                replacements: dict[Name, Name] = {}
                for arg, value in zip(definition.parameters, node.arguments):
                    assert isinstance(arg, Name) and isinstance(value, Name)
                    replacements[arg] = value
                ReplaceName(replacements).visit(definition.body)
                node.replace(definition.body)
                definition.remove()
            elif len(var.reads) == 1:
                # Replaces a function call with an inline function definition, and finally a call
                # While this still involves the `function` keyword, it only requires a set of parantheses
                # (function(params)body end)(), instead of a name (necessarily a split from the keyword)
                # and potentially spaces before and after the definition
                function = ExpFunctionDefinition(
                    node.token, definition.parameters, definition.body
                )
                call = ExpFunctionCall(node.token, function, node.arguments)
                function.parent_class = call
                node.replace(call)
                definition.remove()



import unittest

from tumfl.AST import *
from tumfl.AST.base_function_definition import BaseFunctionDefinition
from tumfl.lexer import Lexer, LexerError
from tumfl.parser import Hint, Parser, ParserError
from tumfl.Token import Token, TokenType

EOF_TOKEN: Token = Token(TokenType.EOF, "eof", 0, 0)


class TestParser(unittest.TestCase):
    @staticmethod
    def parse_number(to_parse: str) -> Number:
        lex = Lexer(to_parse)
        return Number.from_token(lex.get_next_token())

    @staticmethod
    def parse_name(name: str) -> Name:
        return Name.from_token(Token(TokenType.NAME, name, 0, 0))

    @staticmethod
    def parse_string(string: str) -> String:
        return String.from_token(Token(TokenType.STRING, string, 0, 0))

    @staticmethod
    def get_chunk(*statements: Statement) -> Chunk:
        return Chunk(statements[0].token, list(statements), None)

    def test_simple_exp(self):
        parser = Parser("1+2")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            self.parse_number("1"),
            self.parse_number("2"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("4+5*6")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            self.parse_number("4"),
            BinOp.from_token(
                Token(TokenType.MULT, "*", 0, 0),
                self.parse_number("5"),
                self.parse_number("6"),
            ),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("true and false or 1")
        expected_tree = BinOp.from_token(
            Token(TokenType.OR, "or", 0, 0),
            BinOp.from_token(
                Token(TokenType.AND, "and", 0, 0),
                Boolean.from_token(Token(TokenType.TRUE, "true", 0, 0)),
                Boolean.from_token(Token(TokenType.FALSE, "false", 0, 0)),
            ),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("-nil")
        expected_tree_u = UnOp.from_token(
            Token(TokenType.MINUS, "-", 0, 0),
            Nil.from_token(Token(TokenType.NIL, "nil", 0, 0)),
        )
        repr_test = repr(expected_tree_u)
        self.assertEqual(parser._parse_exp(), expected_tree_u)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("\n#'a'")
        expected_tree_u = UnOp.from_token(
            Token(TokenType.HASH, "#", 0, 0),
            self.parse_string("a"),
        )
        repr_test = repr(expected_tree_u)
        self.assertEqual(parser._parse_exp(), expected_tree_u)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("2^-2+s")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            BinOp.from_token(
                Token(TokenType.EXPONENT, "#", 0, 0),
                self.parse_number("2"),
                UnOp.from_token(
                    Token(TokenType.MINUS, "-", 0, 0), self.parse_number("2")
                ),
            ),
            self.parse_name("s"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("1+2+3")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            BinOp.from_token(
                Token(TokenType.PLUS, "+", 0, 0),
                self.parse_number("1"),
                self.parse_number("2"),
            ),
            self.parse_number("3"),
        )
        self.assertEqual(parser._parse_exp(), expected_tree)

    def test_associativity(self):
        parser = Parser("1^2^3")
        expected_tree = BinOp.from_token(
            Token(TokenType.EXPONENT, "^", 0, 0),
            self.parse_number("1"),
            BinOp.from_token(
                Token(TokenType.EXPONENT, "^", 0, 0),
                self.parse_number("2"),
                self.parse_number("3"),
            ),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("1+2-3")
        expected_tree = BinOp.from_token(
            Token(TokenType.MINUS, "-", 0, 0),
            BinOp.from_token(
                Token(TokenType.PLUS, "+", 0, 0),
                self.parse_number("1"),
                self.parse_number("2"),
            ),
            self.parse_number("3"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("foo..bar..baz")
        expected_tree = BinOp.from_token(
            Token(TokenType.CONCAT, "..", 0, 0),
            self.parse_name("foo"),
            BinOp.from_token(
                Token(TokenType.CONCAT, "..", 0, 0),
                self.parse_name("bar"),
                self.parse_name("baz"),
            ),
        )
        self.assertEqual(parser._parse_exp(), expected_tree)

    def test_error(self):
        parser = Parser("1+")
        with self.assertRaises(ParserError) as pe:
            parser._parse_exp()
        self.assertEqual(str(pe.exception), "Unexpected expression")

    def test_unary_expr(self):
        parser = Parser("1- -2")
        expected_tree = BinOp.from_token(
            Token(TokenType.MINUS, "-", 0, 0),
            self.parse_number("1"),
            UnOp.from_token(Token(TokenType.MINUS, "-", 0, 0), self.parse_number("2")),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_len_operator(self):
        parser = Parser("\n#a + 1")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            UnOp.from_token(
                Token(TokenType.HASH, "#", 0, 0),
                self.parse_name("a"),
            ),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_wrong_token(self):
        parser = Parser("a,b+")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("a=1\nb,c+")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")

    def test_parse_var(self):
        parser = Parser('a["b"].d')
        expected_tree = NamedIndex(
            Token(TokenType.NAME, "d", 0, 0),
            Index(
                Token(TokenType.R_BRACKET, "[", 0, 0),
                self.parse_name("a"),
                self.parse_string("b"),
            ),
            self.parse_name("d"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_var(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(2)[1]")
        expected_tree_i = Index(
            Token(TokenType.L_BRACKET, "[", 0, 0),
            self.parse_number("2"),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree_i)
        self.assertEqual(parser._parse_var(), expected_tree_i)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(2)(1)")
        expected_tree_f = ExpFunctionCall(
            Token(TokenType.L_BRACKET, "[", 0, 0),
            self.parse_number("2"),
            [self.parse_number("1")],
        )
        repr_test = repr(expected_tree_f)
        self.assertEqual(parser._parse_var(), expected_tree_f)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("((2)(1))")
        expected_tree_pf = ExpFunctionCall(
            Token(TokenType.L_PAREN, "(", 0, 0),
            self.parse_number("2"),
            [self.parse_number("1")],
            has_parentheses=True,
        )
        self.assertEqual(parser._parse_var(), expected_tree_pf)
        parser = Parser("a:b()")
        expected_tree_m = ExpMethodInvocation(
            Token(TokenType.COLON, "[", 0, 0),
            self.parse_name("a"),
            self.parse_name("b"),
            [],
        )
        repr_test = repr(expected_tree_m)
        self.assertEqual(parser._parse_var(), expected_tree_m)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_table_constructor(self):
        lbracket_token = Token(TokenType.L_BRACKET, "[", 0, 0)
        lcurl_token = Token(TokenType.L_CURL, "{", 0, 0)
        parser = Parser("{}")
        expected_tree = Table(lcurl_token, [])
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("{a}")
        expected_tree = Table(
            lcurl_token,
            [
                NumberedTableField(
                    Token(TokenType.NAME, "a", 0, 0), self.parse_name("a")
                )
            ],
        )
        repr_test = repr(expected_tree)
        result = parser._parse_table_constructor()
        self.assertEqual(result, expected_tree)
        self.assertEqual(result.token, expected_tree.token)
        self.assertEqual(result.fields[0].token, expected_tree.fields[0].token)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser('{["b"]=c}')
        expected_tree = Table(
            lcurl_token,
            [
                ExplicitTableField(
                    lbracket_token,
                    self.parse_string("b"),
                    self.parse_name("c"),
                )
            ],
        )
        repr_test = repr(expected_tree)
        result = parser._parse_table_constructor()
        self.assertEqual(result, expected_tree)
        self.assertEqual(result.fields[0].token, expected_tree.fields[0].token)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("{a=b}")
        expected_tree = Table(
            lcurl_token,
            [
                NamedTableField(
                    Token(TokenType.NAME, "a", 0, 0),
                    self.parse_name("a"),
                    self.parse_name("b"),
                )
            ],
        )
        repr_test = repr(expected_tree)
        result = parser._parse_table_constructor()
        self.assertEqual(result, expected_tree)
        self.assertEqual(result.fields[0].token, expected_tree.fields[0].token)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("{a, b,}")
        expected_tree = Table(
            lcurl_token,
            [
                NumberedTableField(
                    Token(TokenType.NAME, "a", 0, 0), self.parse_name("a")
                ),
                NumberedTableField(
                    Token(TokenType.NAME, "b", 0, 0), self.parse_name("b")
                ),
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("{f'alo'..'xixi'}")
        expected_tree = Table(
            lcurl_token,
            [
                NumberedTableField(
                    Token(TokenType.NAME, "f", 0, 0),
                    BinOp.from_token(
                        Token(TokenType.CONCAT, "..", 0, 0),
                        ExpFunctionCall(
                            Token(TokenType.FUNCTION, "function", 0, 0),
                            self.parse_name("f"),
                            [self.parse_string("alo")],
                        ),
                        self.parse_string("xixi"),
                    ),
                )
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("if }")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.IF, "if", 0, 0), "table constructor", "fields")],
        )
        parser = Parser("{")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected token")
        table_fields_hint = Hint(lcurl_token, "table constructor", "fields")
        self.assertEqual(pe.exception.hints, [table_fields_hint])
        parser = Parser("{[if]=1}")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [
                table_fields_hint,
                Hint(lbracket_token, "explicit table field", "key expression"),
            ],
        )
        parser = Parser("{[a if=1}")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [
                table_fields_hint,
                Hint(lbracket_token, "explicit table field", "key expression"),
            ],
        )
        parser = Parser("{[a]if 1}")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [
                table_fields_hint,
                Hint(lbracket_token, "explicit table field", "value expression"),
            ],
        )
        parser = Parser("{[a]=if}")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [
                table_fields_hint,
                Hint(lbracket_token, "explicit table field", "value expression"),
            ],
        )
        parser = Parser("{a=if}")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [
                table_fields_hint,
                Hint(
                    Token(TokenType.NAME, "a", 0, 0),
                    "named table field",
                    "value expression",
                ),
            ],
        )
        parser = Parser("{if}")
        with self.assertRaises(ParserError) as pe:
            parser._parse_table_constructor()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [
                table_fields_hint,
                Hint(
                    Token(TokenType.IF, "if", 0, 0),
                    "numbered table field",
                    "expression",
                ),
            ],
        )

    def test_parse_args(self):
        parser = Parser("()")
        self.assertEqual(parser._parse_args(), [])
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser('(a, b, "d", 1)')
        expected_args = [
            self.parse_name("a"),
            self.parse_name("b"),
            self.parse_string("d"),
            self.parse_number("1"),
        ]
        self.assertEqual(parser._parse_args(), expected_args)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(a,b")
        with self.assertRaises(ParserError) as pe:
            parser._parse_args()
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("(a,b,)")
        with self.assertRaises(ParserError) as pe:
            parser._parse_args()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        parser = Parser('"abc"')
        self.assertEqual(parser._parse_args(), [self.parse_string("abc")])
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser('{a,"b"}')
        table_result = parser._parse_table_constructor()
        parser = Parser('{a,"b"}')
        self.assertEqual(parser._parse_args(), [table_result])
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(1")
        with self.assertRaises(ParserError) as pe:
            parser._parse_args()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.L_PAREN, "(", 0, 0), "function call", "arguments")],
        )

    def test_parse_names(self):
        parser = Parser("a")
        expected_names = [self.parse_name("a")]
        self.assertEqual(parser._parse_name_list(), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a, b")
        expected_names = [self.parse_name("a"), self.parse_name("b")]
        self.assertEqual(parser._parse_name_list(), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("b")
        expected_names = [self.parse_name("a")]
        self.assertEqual(parser._parse_name_list(self.parse_name("a")), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.NAME, "b", 0, 0))
        parser = Parser(",b")
        expected_names = [self.parse_name("a"), self.parse_name("b")]
        self.assertEqual(parser._parse_name_list(self.parse_name("a")), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a,")
        with self.assertRaises(ParserError) as pe:
            parser._parse_name_list()
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("a, 1")
        with self.assertRaises(ParserError) as pe:
            parser._parse_name_list()
        self.assertEqual(str(pe.exception), "Unexpected token")

    def test_parse_vararg(self):
        parser = Parser("...")
        expected_tree = Vararg(Token(TokenType.ELLIPSIS, "...", 0, 0))
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(...)")
        expected_tree = Vararg(
            Token(TokenType.ELLIPSIS, "...", 0, 0), has_parentheses=True
        )
        self.assertEqual(parser._parse_exp(), expected_tree)

    def test_parse_table_expr(self):
        parser = Parser("{a}")
        expected_tree = Parser("{a}")._parse_table_constructor()
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_parse_funcbody(self):
        function_token = Token(TokenType.FUNCTION, "function", 0, 0)
        parser = Parser("()end")
        expected_tree = BaseFunctionDefinition(
            function_token, [], Block(Token(TokenType.END, "end", 0, 0), [], None)
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(a,b)end")
        expected_tree = BaseFunctionDefinition(
            function_token,
            [self.parse_name("a"), self.parse_name("b")],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(...)end")
        expected_tree = BaseFunctionDefinition(
            function_token,
            [Vararg.from_token(Token(TokenType.ELLIPSIS, "...", 0, 0))],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("(a,...)end")
        expected_tree = BaseFunctionDefinition(
            function_token,
            [
                self.parse_name("a"),
                Vararg.from_token(Token(TokenType.ELLIPSIS, "...", 0, 0)),
            ],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("function a if )end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(function_token, "function", "parameters")]
        )
        parser = Parser("function a(if end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(function_token, "function", "body")])

    def test_parse_func_expr(self):
        function_token = Token(TokenType.FUNCTION, "function", 0, 0)
        parser = Parser("function()end")
        expected_tree = ExpFunctionDefinition(
            function_token,
            [],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("function(if)end")
        with self.assertRaises(ParserError) as pe:
            parser._parse_exp()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(function_token, "function expression", "body")]
        )

    def test_assign(self):
        parser = Parser('a = "bcd"')
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.NAME, "a", 0, 0),
                [self.parse_name("a")],
                [self.parse_string("bcd")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser('a,b,c = "bcd"')
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.NAME, "a", 0, 0),
                [
                    self.parse_name("a"),
                    self.parse_name("b"),
                    self.parse_name("c"),
                ],
                [self.parse_string("bcd")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser('a = "bcd", "def"')
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.NAME, "a", 0, 0),
                [self.parse_name("a")],
                [
                    self.parse_string("bcd"),
                    self.parse_string("def"),
                ],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a, if = b")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected variable")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.COMMA, ",", 0, 0), "assignment", "variables")],
        )
        parser = Parser("(if) = 1")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.L_PAREN, "(", 0, 0), "expression var", "expression")],
        )
        parser = Parser("(foo if = 1")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.L_PAREN, "(", 0, 0), "expression var", "expression")],
        )

    def test_function_stmt(self):
        parser = Parser("a()")
        expected_tree = self.get_chunk(
            FunctionCall(Token(TokenType.NAME, "a", 0, 0), self.parse_name("a"), [])
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a[1].b()")
        expected_tree = self.get_chunk(
            FunctionCall(
                Token(TokenType.NAME, "a", 0, 0),
                NamedIndex(
                    Token(TokenType.NAME, "b", 0, 0),
                    Index(
                        Token(TokenType.L_BRACKET, "[", 0, 0),
                        self.parse_name("a"),
                        self.parse_number("1"),
                    ),
                    self.parse_name("b"),
                ),
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_method_stmt(self):
        parser = Parser("a:b()")
        expected_tree = self.get_chunk(
            MethodInvocation(
                Token(TokenType.NAME, "a", 0, 0),
                self.parse_name("a"),
                self.parse_name("b"),
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_method_expr(self):
        parser = Parser("a:b(1,2)")
        expected_tree = ExpMethodInvocation(
            Token(TokenType.NAME, "a", 0, 0),
            self.parse_name("a"),
            self.parse_name("b"),
            [self.parse_number("1"), self.parse_number("2")],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a:if()")
        with self.assertRaises(ParserError) as pe:
            parser._parse_exp()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.COLON, ":", 0, 0), "invocation", "name")],
        )
        parser = Parser("(a:b(1))")
        expected_tree = ExpMethodInvocation(
            Token(TokenType.NAME, "a", 0, 0),
            self.parse_name("a"),
            self.parse_name("b"),
            [self.parse_number("1")],
            has_parentheses=True,
        )
        self.assertEqual(parser._parse_exp(), expected_tree)

    def test_index_expr(self):
        parser = Parser("a[1]")
        expected_tree = Index(
            Token(TokenType.L_BRACKET, "[", 0, 0),
            self.parse_name("a"),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree)
        result = parser._parse_exp()
        self.assertEqual(result, expected_tree)
        self.assertEqual(result.token, expected_tree.token)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a[if]")
        with self.assertRaises(ParserError) as pe:
            parser._parse_exp()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.L_BRACKET, "[", 0, 0), "index", "expression")],
        )
        parser = Parser("a[1 if")
        with self.assertRaises(ParserError) as pe:
            parser._parse_exp()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.L_BRACKET, "[", 0, 0), "index", "expression")],
        )

    def test_named_index_expr(self):
        parser = Parser("a.b")
        expected_tree = NamedIndex(
            Token(TokenType.DOT, ".", 0, 0),
            self.parse_name("a"),
            self.parse_name("b"),
        )
        repr_test = repr(expected_tree)
        result = parser._parse_exp()
        self.assertEqual(result, expected_tree)
        self.assertEqual(result.token, expected_tree.token)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("a.if")
        with self.assertRaises(ParserError) as pe:
            parser._parse_exp()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(Token(TokenType.DOT, ".", 0, 0), "index", "name")]
        )

    def test_semicolon(self):
        parser = Parser(";")
        expected_tree = self.get_chunk(Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0)))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("foo")
        with self.assertRaises(ParserError) as pe:
            parser._parse_semi()
        self.assertEqual(str(pe.exception), "Unexpected token")

    def test_parse_local_assignment(self):
        parser = Parser("local a, b")
        expected_tree = self.get_chunk(
            LocalAssign(
                Token(TokenType.LOCAL, "local", 0, 0),
                [
                    AttributedName(self.parse_name("a")),
                    AttributedName(self.parse_name("b")),
                ],
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("local a <const>, b <final>")
        expected_tree = self.get_chunk(
            LocalAssign(
                Token(TokenType.LOCAL, "local", 0, 0),
                [
                    AttributedName(self.parse_name("a"), self.parse_name("const")),
                    AttributedName(self.parse_name("b"), self.parse_name("final")),
                ],
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("local a=d,e")
        expected_tree = self.get_chunk(
            LocalAssign(
                Token(TokenType.LOCAL, "local", 0, 0),
                [AttributedName(self.parse_name("a"))],
                [self.parse_name("d"), self.parse_name("e")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("if a = 2")
        with self.assertRaises(ParserError) as pe:
            parser._parse_local()
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("if = 1")
        with self.assertRaises(ParserError) as pe:
            parser._parse_local_assignment(Token(TokenType.LOCAL, "local", 0, 0))
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.IF, "if", 0, 0), "local assign", "names")],
        )
        parser = Parser("local a, if = 1")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.NAME, "a", 0, 0), "local assign", "name")],
        )
        parser = Parser("local a <if> = 1")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.NAME, "a", 0, 0), "local assign", "name attribute")],
        )
        parser = Parser("local a <a< = 1")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.NAME, "a", 0, 0), "local assign", "name attribute")],
        )
        parser = Parser("local a = if")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.NAME, "a", 0, 0), "local assign", "expressions")],
        )

    def test_parse_local_function(self):
        parser = Parser("local function a ()end")
        expected_tree = self.get_chunk(
            LocalFunctionDefinition(
                Token(TokenType.LOCAL, "local", 0, 0),
                self.parse_name("a"),
                [],
                Block(Token(TokenType.END, "end", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("if a()end")
        with self.assertRaises(ParserError) as pe:
            parser._parse_local_function(Token(TokenType.LOCAL, "local", 0, 0))
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.IF, "if", 0, 0), "local function", "name")],
        )

    def test_parse_unknown_local(self):
        parser = Parser("local and")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected symbol after local")

    def test_parse_function(self):
        function_token = Token(TokenType.FUNCTION, "function", 0, 0)
        parser = Parser("function a ()end")
        expected_tree = self.get_chunk(
            FunctionDefinition(
                function_token,
                [self.parse_name("a")],
                None,
                [],
                Block(Token(TokenType.END, "end", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("function a.b:c ()end")
        expected_tree = self.get_chunk(
            FunctionDefinition(
                function_token,
                [self.parse_name("a"), self.parse_name("b")],
                self.parse_name("c"),
                [],
                Block(Token(TokenType.END, "end", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("foo a()end")
        with self.assertRaises(ParserError) as pe:
            parser._parse_function()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.NAME, "foo", 0, 0), "function", "name")],
        )
        parser = Parser("function a:if()end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(function_token, "function", "method name")]
        )

    def test_parse_iterative_for(self):
        for_token = Token(TokenType.FOR, "for", 0, 0)
        parser = Parser("for a in 2 do end")
        expected_tree = self.get_chunk(
            IterativeFor(
                for_token,
                [self.parse_name("a")],
                [self.parse_number("2")],
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("for a, b in 1, 2 do end")
        expected_tree = self.get_chunk(
            IterativeFor(
                for_token,
                [self.parse_name("a"), self.parse_name("b")],
                [self.parse_number("1"), self.parse_number("2")],
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("foo")
        with self.assertRaises(ParserError) as pe:
            parser._parse_for()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(parser.current_token, "for", "name")]
        )
        in_token = Token(TokenType.IN, "in", 0, 0)
        comma_token = Token(TokenType.COMMA, ",", 0, 0)
        parser = Parser("for a, if in a do end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(comma_token, "iterative for", "name list")]
        )
        parser = Parser("for a in if do end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints, [Hint(in_token, "iterative for", "expression list")]
        )
        parser = Parser("if 1 if do end")
        with self.assertRaises(ParserError) as pe:
            parser._parse_iterative_for(for_token, self.parse_name("a"))
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.IF, "if", 0, 0), "iterative for", "expression list")],
        )
        parser = Parser("for a in 1 if end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(in_token, "iterative for", "block")])
        parser = Parser("for a,... in 1 do end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints,
            [Hint(Token(TokenType.COMMA, ",", 0, 0), "iterative for", "name list")],
        )

    def test_parse_numeric_for(self):
        parser = Parser("for a=1,b do end")
        for_token = Token(TokenType.FOR, "for", 0, 0)
        expected_tree = self.get_chunk(
            NumericFor(
                for_token,
                self.parse_name("a"),
                self.parse_number("1"),
                self.parse_name("b"),
                None,
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("for a=1,b,3 do end")
        expected_tree = self.get_chunk(
            NumericFor(
                for_token,
                self.parse_name("a"),
                self.parse_number("1"),
                self.parse_name("b"),
                self.parse_number("3"),
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("for foo = end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        assign_token = Token(TokenType.ASSIGN, "=", 0, 0)
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints, [Hint(assign_token, "numeric for", "start expression")]
        )
        parser = Parser("for 1,1 do end")
        with self.assertRaises(ParserError) as pe:
            parser._parse_numeric_for(for_token, self.parse_name("a"))
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("for a = 1,end do end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(
            pe.exception.hints, [Hint(assign_token, "numeric for", "stop expression")]
        )
        self.assertEqual(str(pe.exception), "Unexpected expression")
        parser = Parser("for a = 1 if 2 do end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(
            pe.exception.hints, [Hint(assign_token, "numeric for", "stop expression")]
        )
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("for a = 1,2,end do end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(
            pe.exception.hints, [Hint(assign_token, "numeric for", "step expression")]
        )
        self.assertEqual(str(pe.exception), "Unexpected expression")
        parser = Parser("for a = 1,2 if end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(
            pe.exception.hints, [Hint(assign_token, "numeric for", "block")]
        )
        self.assertEqual(str(pe.exception), "Unexpected token")

    def test_parse_unknown_for(self):
        parser = Parser("for name and")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected for condition")

    def test_parse_if(self):
        parser = Parser("if 1 then end")
        if_token = Token(TokenType.IF, "if", 0, 0)
        then_token = Token(TokenType.THEN, "then", 0, 0)
        else_token = Token(TokenType.ELSE, "else", 0, 0)
        eof_token = EOF_TOKEN
        expected_tree = self.get_chunk(
            If(
                if_token,
                self.parse_number("1"),
                Block(then_token, [], None),
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, eof_token)
        parser = Parser("if 1 then else end")
        expected_tree = self.get_chunk(
            If(
                if_token,
                self.parse_number("1"),
                Block(then_token, [], None),
                Block(else_token, [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, eof_token)
        parser = Parser("if 1 then elseif 2 then elseif 3 then else end")
        expected_tree = self.get_chunk(
            If(
                if_token,
                self.parse_number("1"),
                Block(then_token, [], None),
                If(
                    Token(TokenType.ELSEIF, "elseif", 0, 0),
                    self.parse_number("2"),
                    Block(Token(TokenType.ELSEIF, "elseif", 0, 0), [], None),
                    If(
                        Token(TokenType.ELSEIF, "elseif", 0, 0),
                        self.parse_number("3"),
                        Block(Token(TokenType.ELSEIF, "elseif", 0, 0), [], None),
                        Block(else_token, [], None),
                    ),
                ),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, eof_token)
        parser = Parser("if then")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(pe.exception.hints, [Hint(if_token, "if", "if condition")])
        parser = Parser("if 1 foo end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(if_token, "if", "if block")])
        parser = Parser("if 1 then elseif end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(pe.exception.hints, [Hint(if_token, "if", "elseif condition")])
        parser = Parser("if 1 then elseif 2 foo end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(if_token, "if", "elseif block")])
        parser = Parser("if 1 then elseif 2 then else until")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(if_token, "if", "else block")])
        parser = Parser("foo 1 then end")
        with self.assertRaises(ParserError):
            parser._parse_if()

    def test_parse_repeat(self):
        parser = Parser("repeat until 1")
        repeat_token = Token(TokenType.REPEAT, "repeat", 0, 0)
        expected_tree = self.get_chunk(
            Repeat(
                repeat_token,
                self.parse_number("1"),
                Block(repeat_token, [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("repeat foo")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected lonely name")
        self.assertEqual(pe.exception.hints, [Hint(repeat_token, "repeat", "block")])
        parser = Parser("repeat end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(
            pe.exception.hints, [Hint(repeat_token, "repeat", "condition")]
        )
        parser = Parser("repeat until")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(
            pe.exception.hints, [Hint(repeat_token, "repeat", "condition")]
        )
        parser = Parser("foo until 1 end")
        with self.assertRaises(ParserError):
            parser._parse_repeat()

    def test_parse_while(self):
        parser = Parser("while 1 do end")
        while_token = Token(TokenType.WHILE, "while", 0, 0)
        expected_tree = self.get_chunk(
            While(
                while_token,
                self.parse_number("1"),
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("while")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected expression")
        self.assertEqual(pe.exception.hints, [Hint(while_token, "while", "condition")])
        parser = Parser("while 1 foo end")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        parser = Parser("while 1 do foo")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected lonely name")
        self.assertEqual(pe.exception.hints, [Hint(while_token, "while", "block")])
        parser = Parser("foo 1 do end")
        with self.assertRaises(ParserError):
            parser._parse_while()

    def test_parse_label(self):
        parser = Parser("::label::")
        label_token = Token(TokenType.LABEL_BORDER, "::", 0, 0)
        expected_tree = self.get_chunk(Label(label_token, self.parse_name("label")))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("::")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(label_token, "label", "name")])
        parser = Parser("::label")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(label_token, "label", "end")])
        parser = Parser("foo name::")
        with self.assertRaises(ParserError):
            parser._parse_label()

    def test_parse_goto(self):
        parser = Parser("goto label")
        goto_token = Token(TokenType.GOTO, "goto", 0, 0)
        expected_tree = self.get_chunk(Goto(goto_token, self.parse_name("label")))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("goto")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")
        self.assertEqual(pe.exception.hints, [Hint(goto_token, "goto", "name")])
        parser = Parser("foo label")
        with self.assertRaises(ParserError):
            parser._parse_goto()

    def test_parse_break(self):
        parser = Parser("break")
        expected_tree = self.get_chunk(Break(Token(TokenType.BREAK, "break", 0, 0)))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("foo")
        with self.assertRaises(ParserError):
            parser._parse_break()

    def test_parse_block(self):
        parser = Parser("do end")
        expected_tree = self.get_chunk(Block(Token(TokenType.DO, "do", 0, 0), [], None))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("do ; end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0))],
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("do ;return a; end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0))],
                [self.parse_name("a")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("do ;return a end")
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("do ;return ; end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0))],
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("do return end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [],
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)
        parser = Parser("do")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected token")

    def test_unexpected_statement(self):
        parser = Parser("+")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected statement")

    def test_weird_stuff(self):
        parser = Parser("assert()()")
        expected_tree = self.get_chunk(
            FunctionCall(
                Token(TokenType.NAME, "assert", 0, 0),
                ExpFunctionCall(
                    Token(TokenType.NAME, "assert", 0, 0), self.parse_name("assert"), []
                ),
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, EOF_TOKEN)

    def test_lonely_var(self):
        parser = Parser("a")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        self.assertEqual(str(pe.exception), "Unexpected lonely name")

    def test_lonely_return(self):
        parser = Parser("do return end")
        expected_tree = self.get_chunk(Block(Token(TokenType.DO, "do", 0, 0), [], []))
        self.assertEqual(parser.parse_chunk(), expected_tree)
        parser = Parser("do end")
        expected_tree = self.get_chunk(Block(Token(TokenType.DO, "do", 0, 0), [], None))
        self.assertEqual(parser.parse_chunk(), expected_tree)

    def test_accepts_as(self):
        parser = Parser("as = 'foo'")
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.ASSIGN, "=", 0, 0),
                [self.parse_name("as")],
                [self.parse_string("foo")],
            )
        )
        self.assertEqual(parser.parse_chunk(), expected_tree)
        parser = Parser("as = 'foo'", typed=True)
        with self.assertRaises(ParserError):
            parser.parse_chunk()

    def test_unicode_errors(self):
        parser = Parser("a = '\\xff'")
        with self.assertRaises(LexerError):
            parser.parse_chunk()
        parser = Parser("a = '\\xff'", ignore_unicode_errors=True)
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.ASSIGN, "=", 0, 0),
                [self.parse_name("a")],
                [self.parse_string("")],
            )
        )
        self.assertEqual(parser.parse_chunk(), expected_tree)

    def test_error_message(self):
        parser = Parser("foo()\nbar()\na = ()\nbaz()\n")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        e = pe.exception
        self.assertMultiLineEqual(
            e.full_error,
            "Error on line 3:\nbar()\na = ()\n     ^\nUnexpected expression\nhints: <assignment> expressions (3:3) -> <expression var> expression (3:5)\n",
        )
        self.assertEqual(str(e), "Unexpected expression")
        self.assertEqual(len(e.hints), 2)
        self.assertEqual(e.token, Token(TokenType.R_PAREN, ")", 0, 0))
        parser = Parser("a = ()\nbaz()\n")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        e = pe.exception
        self.assertMultiLineEqual(
            e.full_error,
            "Error on line 1:\na = ()\n     ^\nUnexpected expression\nhints: <assignment> expressions (1:3) -> <expression var> expression (1:5)\n",
        )
        parser = Parser("bar()\na = ()\n")
        with self.assertRaises(ParserError) as pe:
            parser.parse_chunk()
        e = pe.exception
        self.assertMultiLineEqual(
            e.full_error,
            "Error on line 2:\nbar()\na = ()\n     ^\nUnexpected expression\nhints: <assignment> expressions (2:3) -> <expression var> expression (2:5)\n",
        )

    def test_hint(self):
        hint_a = Hint(
            Token(TokenType.NAME, "a", 0, 0), "test hint", "test hint description"
        )
        hint_b = Hint(
            Token(TokenType.NAME, "b", 0, 0), "test hint b", "test hint description b"
        )
        self.assertNotEqual(hint_a, hint_b)
        self.assertNotEqual(hint_a, True)
        self.assertEqual(
            hint_a,
            Hint(
                Token(TokenType.NAME, "a", 0, 0), "test hint", "test hint description"
            ),
        )
        self.assertNotEqual(
            hint_a,
            Hint(
                Token(TokenType.NAME, "a", 0, 0), "test hint", "different description"
            ),
        )
        self.assertNotEqual(
            hint_a,
            Hint(
                Token(TokenType.NAME, "b", 0, 0), "test hint", "test hint description"
            ),
        )
        self.assertNotEqual(
            hint_a,
            Hint(
                Token(TokenType.NAME, "a", 0, 0), "test hint b", "test hint description"
            ),
        )
        self.assertEqual(repr(hint_a), "<test hint> test hint description (0:0)")


class EmptyTest(unittest.TestCase):
    def test_empty(self):
        parser = Parser("")
        parser.parse_chunk()



import unittest

from tumfl import parse
from tumfl.AST import Assign, Chunk
from tumfl.to_lua_type import Retype, to_lua_type
from tumfl.to_python_type import NilVal


class TestToLuaType(unittest.TestCase):
    def run_test(self, python_type: Retype, expected: str):
        expected_ast = parse(f"a = {expected}")
        assert isinstance(expected_ast, Chunk)
        first_statement = expected_ast.statements[0]
        assert isinstance(first_statement, Assign)
        expected_value = first_statement.expressions[0]
        result = to_lua_type(python_type)
        self.assertEqual(result, expected_value)

    def test_string(self):
        self.run_test("hello", '"hello"')

    def test_integer(self):
        self.run_test(1, "1")
        self.run_test(-50, "-50")

    def test_float(self):
        self.run_test(1.0, "1.0")
        self.run_test(-50.5, "-50.5")
        self.run_test(1.23, "1.23")
        self.run_test(-50.5, "-50.5")

    def test_nil(self):
        self.run_test(NilVal(), "nil")

    def test_boolean(self):
        self.run_test(True, "true")
        self.run_test(False, "false")

    def test_list(self):
        self.run_test([1, 2, 3], "{1, 2, 3}")
        self.run_test([1, 2, "a"], '{1, 2, "a"}')
        self.run_test([1, 2, NilVal()], "{1, 2, nil}")

    def test_dictionary(self):
        self.run_test({"a": 1, "b": 2}, "{a = 1, b = 2}")
        self.run_test({1: 1, 2: 2}, "{1, 2}")
        self.run_test({0: 1, 1: 2}, "{2, [0] = 1}")
        self.run_test({0: NilVal(), "a": NilVal()}, "{[0] = nil, a = nil}")
        self.run_test({"1foo": 2}, '{["1foo"] = 2}')


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import Optional

from tumfl.AST import Name


class Replacements:
    def __init__(self, targets: list[Name], original: Optional[Name]):
        self.targets: list[Name] = targets
        for target in targets:
            target.set_attribute(self)
        self.original: Optional[Name] = original
        self.parent: Optional[Replacements] = None
        self.replacement: Optional[str] = None
        self.after: Optional[Name] = None

    def value(self) -> int:
        return len(self.targets)

    def __repr__(self) -> str:
        return f"Replacements({self.targets=}, {self.original=})"


ReplacementCollection = list[list[Replacements]]


def key_function(replacements: list[Replacements]) -> int:
    return sum(replacement.value() for replacement in replacements)



from logging import warning
from pathlib import Path

from .AST import Assign, ASTNode, Name, String
from .basic_walker import NoneWalker
from .dependency_resolver import resolve_recursive


class Config(NoneWalker):
    def __init__(self, replacements: dict[str, ASTNode], prefix: str = "$$"):
        super().__init__()
        self.replacements: dict[str, ASTNode] = replacements
        self.prefix: str = prefix

    def visit_String(self, node: String) -> None:
        if node.value.startswith(self.prefix):
            if node.value[len(self.prefix) :] in self.replacements:
                node.replace(self.replacements[node.value[len(self.prefix) :]])
            else:
                warning(f"Found placeholder with no assigned value: {node.value}")


class ParameterGetter(NoneWalker):
    def __init__(self) -> None:
        super().__init__()
        self.parameters: dict[str, ASTNode] = {}

    def visit_Assign(self, node: Assign) -> None:
        if (
            len(node.targets) == 1
            and isinstance(node.targets[0], Name)
            and len(node.expressions) == 1
        ):
            name = node.targets[0]
            assert isinstance(name, Name)
            self.parameters[name.variable_name] = node.expressions[0]
        else:
            warning(f"Ignoring config parameter {node}")


def parse_config(file: Path) -> Config:
    chunk: ASTNode = resolve_recursive(file, [file.parent])
    getter = ParameterGetter()
    getter.visit(chunk)
    return Config(getter.parameters)



from __future__ import annotations

from typing import NoReturn

from .AST import (
    ExpFunctionDefinition,
    FunctionDefinition,
    LocalFunctionDefinition,
    Name,
    Vararg,
)
from .basic_walker import NoneWalker
from .error import SemanticError
from .Token import Token


class SemanticAnalyzer(NoneWalker):
    def __init__(self) -> None:
        self.in_vararg_funct: bool = False

    def error(self, message: str, token: Token) -> NoReturn:
        raise SemanticError(message, token)

    @staticmethod
    def _param_contains_vararg(parameters: list[Name | Vararg]) -> bool:
        return any(isinstance(param, Vararg) for param in parameters)

    def visit_FunctionDefinition(self, node: FunctionDefinition) -> None:
        before: bool = self.in_vararg_funct
        self.in_vararg_funct = before or self._param_contains_vararg(node.parameters)
        super().visit_FunctionDefinition(node)
        self.in_vararg_funct = before

    def visit_ExpFunctionDefinition(self, node: ExpFunctionDefinition) -> None:
        before: bool = self.in_vararg_funct
        self.in_vararg_funct = before or self._param_contains_vararg(node.parameters)
        super().visit_ExpFunctionDefinition(node)
        self.in_vararg_funct = before

    def visit_LocalFunctionDefinition(self, node: LocalFunctionDefinition) -> None:
        before: bool = self.in_vararg_funct
        self.in_vararg_funct = before or self._param_contains_vararg(node.parameters)
        super().visit_LocalFunctionDefinition(node)
        self.in_vararg_funct = before

    def visit_Vararg(self, node: Vararg) -> None:
        if not self.in_vararg_funct:
            self.error("Vararg outside of vararg function", node.token)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import BaseFunctionDefinition, Block, Name, Vararg
    from tumfl.Token import Token


class FunctionDefinition(Statement):
    """Definition of a lua function, NOT local"""

    def __init__(
        self,
        token: Token,
        names: Sequence[Name],
        method_name: Optional[Name],
        parameters: Sequence[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "FunctionDefinition")
        self.names: list[Name] = list(names)
        self.method_name: Optional[Name] = method_name
        self.parameters: list[Name | Vararg] = list(parameters)
        self.body: Block = body

    @staticmethod
    def from_base_definition(
        base: BaseFunctionDefinition,
        names: list[Name],
        method_name: Optional[Name],
    ) -> FunctionDefinition:
        return FunctionDefinition(
            base.token, names, method_name, base.parameters, base.body
        )



from __future__ import annotations

from typing import Any, Optional

from tumfl.AST import Name, NamedIndex
from tumfl.minifier.util import MutableBool, SharedList
from tumfl.minifier.util.replacements import ReplacementCollection, Replacements


class Variable:
    def __init__(self, is_global: bool, preserve: bool) -> None:
        # ideally, we would use sets, but they compare by equality, and all names are equal
        self.writes: list[Name] = []
        self.reads: list[Name] = []
        self.children: dict[str, Variable] = {}
        self.scopes: list[tuple[Scope | Variable, Name]] = []
        self.has_variable_access: MutableBool = MutableBool(False)
        self.is_global: MutableBool = MutableBool(is_global)
        self.preserve: MutableBool = MutableBool(preserve)
        self._shared: SharedList[Variable] = SharedList(self)

    def remove(self, name: Name) -> None:
        if name not in self:
            return
        for i, test in enumerate(self.writes):
            if test is name:
                self.writes.pop(i)
                break
        else:
            for i, test in enumerate(self.reads):
                if test is name:
                    self.reads.pop(i)
                    break
            else:
                return
        parent = name.parent_class
        if isinstance(parent, NamedIndex) and name is parent.variable_name:
            for i, other_scope in enumerate(self.scopes):
                if name is other_scope[1]:
                    for own in self.reads + self.writes:
                        if own == name and own is not name:
                            self.scopes[i] = (other_scope[0], own)
                            break

        self.cleanup()

    def add_scope(self, parent: Variable | Scope, name: Name) -> None:
        if (parent, name) not in self.scopes:
            self.scopes.append((parent, name))

    def cleanup(self) -> None:
        if len(self.writes) == 0 and len(self.reads) == 0:
            for scope, name in self.scopes:
                scope.remove_var(name.variable_name)
            self.scopes.clear()

    def __register(self, name: Name) -> None:
        name.set_attribute(self)
        if name.parent_class:
            name.parent_class.set_attribute_if_not_exists(self)

    def add_write(self, write: Name) -> None:
        self.writes.append(write)
        self.__register(write)

    def __contains__(self, item: Any) -> bool:
        # in does not work, as all names are equal
        if any(write is item for write in self.writes):
            return True
        return any(read is item for read in self.reads)

    def add_read(self, read: Name, is_method: bool) -> None:
        if read not in self:
            self.reads.append(read)
            self.__register(read)
        if is_method:
            # do not mess with method invocations
            self.preserve.value = True

    def add_variable(self, name: Name) -> Variable:
        self.children[name.variable_name] = self.children.get(
            name.variable_name, Variable(self.is_global.value, self.preserve.value)
        )
        self.children[name.variable_name].add_scope(self, name)
        return self.children[name.variable_name]

    def remove_var(self, name: str) -> None:
        del self.children[name]

    def __repr__(self) -> str:
        return f"Variable({self.writes=}, {self.reads=}, {self.children=}, {self.has_variable_access=}, {self.is_global=})"

    def is_unused(self) -> bool:
        return len(self.reads) == 0 and all(
            child.is_unused() for child in self.children.values()
        )

    def merge(self, other: Variable) -> None:
        self.writes.extend(other.writes)
        self.reads.extend(other.reads)
        for name, children in other.children.items():
            if name in self.children:
                self.children[name].merge(children)
            else:
                self.children[name] = children
        self.scopes.extend(other.scopes)
        _side_effect = self.has_variable_access or other.has_variable_access
        _side_effect = self.is_global or other.is_global
        _side_effect = self.preserve or other.preserve  # noqa: F841
        # pylint: disable=protected-access
        for obj in self._shared.absorb(other._shared):
            obj.writes = self.writes
            obj.reads = self.reads
            obj.scopes = self.scopes
            obj.children = self.children
            obj.has_variable_access = self.has_variable_access
            obj.is_global = self.is_global
            obj.preserve = self.preserve

    def collect_replacements(self) -> ReplacementCollection:
        if self.preserve:
            return []
        replacements: ReplacementCollection = []
        self_value: int = len(self.writes) + len(self.reads)
        if not self.has_variable_access.value and len(self.writes) <= 1:
            for children in self.children.values():
                new_replacements = children.collect_replacements()
                self_value -= (
                    sum(
                        len(reads.targets)
                        for replacements in new_replacements
                        for reads in replacements
                    )
                    - 1
                )
                replacements.extend(new_replacements)
        alias: Optional[Name] = None
        do_replacement: bool = False
        # we can only recover cost-1 from aliases, minus 4 for new variable, =, and ;
        self_acquired_space = (self_value - 1) * len(
            self.scopes[0][1].variable_name
        ) - 4
        if len(self.writes) == 0 and self.is_global.value and self_acquired_space > 0:
            # is a global variable, declare an alias
            alias = self.scopes[0][1]
            do_replacement = True
        elif len(self.writes) > 0 and all(
            isinstance(scope[0], Scope) for scope in self.scopes
        ):
            # is a standard variable, simply rename it
            do_replacement = True
        if do_replacement:
            new_replacement = Replacements(self.reads + self.writes, alias)
            for inner_list in replacements:
                for inner in inner_list:
                    if inner.parent is None:
                        inner.parent = new_replacement
                    if self.writes and inner.after is None:
                        inner.after = self.writes[0]
            replacements.append([new_replacement])
        return replacements


# circular dependency
# pylint: disable=wrong-import-position
from tumfl.minifier.util.scope import Scope



from __future__ import annotations

from typing import TYPE_CHECKING

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import Name
    from tumfl.Token import Token


class Goto(Statement):
    """The goto statement"""

    def __init__(self, token: Token, label_name: Name):
        super().__init__(token, "Goto")
        self.label_name: Name = label_name



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from tumfl.AST import (
    Assign,
    BinaryOperand,
    BinOp,
    Block,
    Boolean,
    Break,
    ExpFunctionCall,
    ExplicitTableField,
    Goto,
    Index,
    Label,
    LocalAssign,
    Name,
    NamedTableField,
    Nil,
    Number,
    NumberedTableField,
    Semicolon,
    String,
    Table,
    UnaryOperand,
    UnOp,
    Vararg,
)
from tumfl.basic_walker import BasicWalker


class NilVal:
    """Pseudo class that defines an nil value, in order to differentiate it from None."""

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, NilVal)


if TYPE_CHECKING:
    Retype = Union[
        list["Retype"], dict["Retype", "Retype"], str, int, float, bool, None, NilVal
    ]
else:
    Retype = object


def to_bool(val: Retype) -> bool:
    return not isinstance(val, NilVal) and val is not False


class ToPythonType(BasicWalker[Retype]):
    # pylint: disable=duplicate-code
    def visit_Name(self, node: Name) -> Retype:
        pass

    def visit_Vararg(self, node: Vararg) -> Retype:
        pass

    def visit_Assign(self, node: Assign) -> Retype:
        pass

    def visit_Block(self, node: Block) -> Retype:
        pass

    def visit_Break(self, node: Break) -> Retype:
        pass

    def visit_Goto(self, node: Goto) -> Retype:
        pass

    def visit_Label(self, node: Label) -> Retype:
        pass

    def visit_LocalAssign(self, node: LocalAssign) -> Retype:
        pass

    def visit_Semicolon(self, node: Semicolon) -> Retype:
        pass

    def visit_Index(self, node: Index) -> Retype:
        pass

    def visit_ExpFunctionCall(self, node: ExpFunctionCall) -> Retype:
        pass

    def visit_Boolean(self, node: Boolean) -> Retype:
        return node.value

    def visit_Nil(self, node: Nil) -> Retype:
        return NilVal()

    def visit_Number(self, node: Number) -> Retype:
        int_repr: Optional[int] = node.to_int()
        return int_repr if int_repr is not None else node.to_float()

    def visit_String(self, node: String) -> Retype:
        return node.value

    def visit_Table(self, node: Table) -> Retype:
        results: dict[Retype, Retype] = {}
        index: int = 1
        for field in node.fields:
            field_value = self.visit(field.value)
            if field_value is None:
                return None
            if isinstance(field, NamedTableField):
                results[field.field_name.variable_name] = field_value
            elif isinstance(field, NumberedTableField):
                results[index] = field_value
                index += 1
            elif isinstance(field, ExplicitTableField):
                key_value = self.visit(field.at)
                if key_value is None:
                    return None
                results[key_value] = field_value
            else:
                raise NotImplementedError
        # pylint: disable=consider-iterating-dictionary
        if len(results) > 0 and all(isinstance(val, int) for val in results.keys()):
            max_key: int = max(results.keys())  # type: ignore
            min_key: int = min(results.keys())  # type: ignore
            if min_key == 1 and all(key in results for key in range(1, max_key + 1)):
                return [results[key] for key in range(1, max_key + 1)]
        return results

    def visit_UnOp(self, node: UnOp) -> Retype:
        rhs = self.visit(node.right)
        if rhs is None:
            return None
        if node.op == UnaryOperand.NOT:
            return not to_bool(rhs)
        if node.op == UnaryOperand.MINUS and isinstance(rhs, (int, float)):
            return -rhs
        if node.op == UnaryOperand.HASH and isinstance(rhs, (list, dict, str)):
            return len(rhs)
        if node.op == UnaryOperand.BIT_XOR and isinstance(rhs, (int, float)):
            return ~int(rhs)
        return None

    def visit_BinOp(self, node: BinOp) -> Retype:
        rhs = self.visit(node.right)
        lhs = self.visit(node.left)
        int_or_float = isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
        str_int_or_float = isinstance(lhs, (str, int, float)) and isinstance(
            rhs, (str, int, float)
        )
        bit_safe = (
            isinstance(lhs, (int, float))
            and isinstance(rhs, (int, float))
            and lhs % 1 == 0
            and rhs % 1 == 0
        )
        if rhs is None or lhs is None:
            return None
        if node.op == BinaryOperand.PLUS and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return lhs + rhs
        if node.op == BinaryOperand.MINUS and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return lhs - rhs
        if node.op == BinaryOperand.MULT and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return lhs * rhs
        if node.op == BinaryOperand.DIVIDE and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return float(lhs) / float(rhs)
        if node.op == BinaryOperand.MODULO and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return lhs % rhs
        if node.op == BinaryOperand.INTEGER_DIVISION and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return int(lhs // rhs)
        if node.op == BinaryOperand.EXPONENT and int_or_float:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return float(lhs) ** float(rhs)
        if node.op == BinaryOperand.EQUALS and str_int_or_float:
            return lhs == rhs
        if node.op == BinaryOperand.NOT_EQUALS and str_int_or_float:
            return lhs != rhs
        if node.op == BinaryOperand.GREATER_THAN and str_int_or_float:
            if isinstance(lhs, str) and isinstance(rhs, str):
                return lhs > rhs
            if isinstance(lhs, (int, float)) and isinstance(rhs, (int, float)):
                return lhs > rhs
        if node.op == BinaryOperand.GREATER_EQUALS and str_int_or_float:
            if isinstance(lhs, str) and isinstance(rhs, str):
                return lhs >= rhs
            if isinstance(lhs, (int, float)) and isinstance(rhs, (int, float)):
                return lhs >= rhs
        if node.op == BinaryOperand.LESS_THAN and str_int_or_float:
            if isinstance(lhs, str) and isinstance(rhs, str):
                return lhs < rhs
            if isinstance(lhs, (int, float)) and isinstance(rhs, (int, float)):
                return lhs < rhs
        if node.op == BinaryOperand.LESS_EQUALS and str_int_or_float:
            if isinstance(lhs, str) and isinstance(rhs, str):
                return lhs <= rhs
            if isinstance(lhs, (int, float)) and isinstance(rhs, (int, float)):
                return lhs <= rhs
        if node.op == BinaryOperand.BIT_AND and bit_safe:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return int(lhs) & int(rhs)
        if node.op == BinaryOperand.BIT_OR and bit_safe:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return int(lhs) | int(rhs)
        if node.op == BinaryOperand.BIT_XOR and bit_safe:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return int(lhs) ^ int(rhs)
        if node.op == BinaryOperand.BIT_SHIFT_LEFT and bit_safe:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return int(lhs) << int(rhs)
        if node.op == BinaryOperand.BIT_SHIFT_RIGHT and bit_safe:
            assert isinstance(lhs, (int, float)) and isinstance(rhs, (int, float))
            return int(lhs) >> int(rhs)
        if node.op == BinaryOperand.CONCAT and str_int_or_float:
            assert isinstance(lhs, (str, int, float)) and isinstance(
                rhs, (str, int, float)
            )
            return str(lhs) + str(rhs)
        if node.op == BinaryOperand.AND:
            return lhs if not to_bool(lhs) else rhs
        if node.op == BinaryOperand.OR:
            return lhs if to_bool(lhs) else rhs
        return None
